/*
 * Being that the windows in TDE are numbered and lettered, we can easily
 *   prompt for windows to diff.  Might as well do a few standard diff
 *   options:  ignore leading space, ignore all space, ignore blank lines,
 *   and Ignore/Match case.  Once the diff is defined, just press one key
 *   to find the next diff.  Any two visible windows may be diffed.
 *
 * Let's add a non-standard diff option for writers -- ignore end of line.
 *   Diffing a paragraph with only margin changes will produce no diffs.
 *
 * New editor name:  TDE, the Thomson-Davis Editor.
 * Author:           Frank Davis
 * Date:             June 5, 1991, version 1.0
 * Date:             July 29, 1991, version 1.1
 * Date:             October 5, 1991, version 1.2
 * Date:             January 20, 1992, version 1.3
 * Date:             February 17, 1992, version 1.4
 * Date:             April 1, 1992, version 1.5
 * Date:             June 5, 1992, version 2.0
 * Date:             October 31, 1992, version 2.1
 *
 * This code is released into the public domain, Frank Davis.
 * You may distribute it freely.
 */

#include "tdestr.h"
#include "common.h"
#include "define.h"
#include "tdefunc.h"


/*
 * Name:    define_diff
 * Purpose: get info needed to initialize diff
 * Date:    October 31, 1992
 * Passed:  window:  pointer to current window
 * Notes:   allow the user to start the diff at the beginning of the
 *            file or at the current cursor location.  once the diff
 *            has been defined the user may press one key to diff again.
 *          user may diff any two visible windows on the screen.
 */
int  define_diff( WINDOW *window )
{
int  rc;
char temp[MAX_COLS];
int  num1;
int  let1;
int  num2;
int  let2;
int  start;
char line_buff[(MAX_COLS+1)*2];  /* buffer for char and attribute  */
char buff[MAX_COLS*2];           /* buffer for char and attribute  */

   /*
    * get window number and letter of the first diff window.  then,
    *   verify that window - does it exit? is it visible?
    */
   *temp = '\0';
   rc = get_name( diff_prompt1, window->bottom_line, temp, g_display.message_color );
   if (rc == OK) {
      rc = verify_number( temp, &num1 );
      if (rc == OK)
         rc = verify_letter( temp, &let1, &diff.w1 );
   } else
      return( ERROR );
   if (rc == ERROR) {
      combine_strings( buff, diff_prompt6a, temp, diff_prompt6b );
      error( WARNING, window->bottom_line, buff );
      return( ERROR );
   }

   /*
    * get and verify the next window number and letter to diff.
    */
   *temp = '\0';
   rc = get_name( diff_prompt2, window->bottom_line, temp, g_display.message_color );
   if (rc == OK) {
      rc = verify_number( temp, &num2 );
      if (rc == OK)
         rc = verify_letter( temp, &let2, &diff.w2 );
   } else
      return( ERROR );
   if (rc == ERROR) {
      combine_strings( buff, diff_prompt6a, temp, diff_prompt6b );
      error( WARNING, window->bottom_line, buff );
      return( ERROR );
   }

   /*
    * are leading spaces significant?
    */
   save_screen_line( 0, window->bottom_line, line_buff );
   set_prompt( diff_prompt7a, window->bottom_line );
   start = get_yn( );
   restore_screen_line( 0, window->bottom_line, line_buff );
   if (start != ERROR) {
      if (start == A_YES)
         diff.leading = TRUE;
      else
         diff.leading = FALSE;
   } else
      return( ERROR );

   /*
    * are all spaces significant?
    */
   save_screen_line( 0, window->bottom_line, line_buff );
   set_prompt( diff_prompt7b, window->bottom_line );
   start = get_yn( );
   restore_screen_line( 0, window->bottom_line, line_buff );
   if (start != ERROR) {
      if (start == A_YES)
         diff.leading = diff.all_space = TRUE;
      else
         diff.all_space = FALSE;
   } else
      return( ERROR );

   /*
    * are blank lines significant?
    */
   save_screen_line( 0, window->bottom_line, line_buff );
   set_prompt( diff_prompt7c, window->bottom_line );
   start = get_yn( );
   restore_screen_line( 0, window->bottom_line, line_buff );
   if (start != ERROR) {
      if (start == A_YES)
         diff.blank_lines = TRUE;
      else
         diff.blank_lines = FALSE;
   } else
      return( ERROR );

   /*
    * is end of line significant?
    */
   save_screen_line( 0, window->bottom_line, line_buff );
   set_prompt( diff_prompt7d, window->bottom_line );
   start = get_yn( );
   restore_screen_line( 0, window->bottom_line, line_buff );
   if (start != ERROR) {
      if (start == A_YES)
         diff.ignore_eol = TRUE;
      else
         diff.ignore_eol = FALSE;
   } else
      return( ERROR );

   /*
    * now, find out were to start the diff -- beginning of file or
    *   current cursor location.
    */
   save_screen_line( 0, window->bottom_line, line_buff );
   set_prompt( diff_prompt3, window->bottom_line );
   start = get_bc( );
   restore_screen_line( 0, window->bottom_line, line_buff );

   if (start != ERROR) {
      entab_linebuff( );
      un_copy_line( window->cursor, window, TRUE );

      /*
       * if everything is everything, initialize the diff pointers.
       */
      diff.defined = TRUE;
      if (start == BEGINNING) {
         diff.d1 = cpf( diff.w1->file_info->start_text );
         diff.d2 = cpf( diff.w2->file_info->start_text );
         rc = differ( 0, 0, window->bottom_line );
      } else {
         diff.d1 = cpf( diff.w1->cursor );
         diff.d2 = cpf( diff.w2->cursor );
         rc = differ( diff.w1->rcol, diff.w2->rcol, window->bottom_line );
      }
   }
   return( rc );
}


/*
 * Name:    repeat_diff
 * Purpose: compare two cursor positions
 * Date:    October 31, 1992
 * Passed:  window:  pointer to current window
 * Notes:   user may press this key at any time once the diff has been
 *            defined.
 */
int  repeat_diff( WINDOW *window )
{
register int rc = ERROR;

   if (diff.defined) {
      entab_linebuff( );
      un_copy_line( window->cursor, window, TRUE );

      /*
       * initialize the diff pointers.
       */
      diff.d1 = cpf( diff.w1->cursor );
      diff.d2 = cpf( diff.w2->cursor );
      rc = differ( diff.w1->rcol, diff.w2->rcol, window->bottom_line );
   } else
      error( WARNING, window->bottom_line, diff_prompt5 );
   return( rc );
}


/*
 * Name:    differ
 * Purpose: diff text pointers
 * Date:    October 31, 1992
 * Passed:  none
 * Notes:   a straight diff on text pointers is simple; however, diffing
 *            with leading spaces and tabs is kinda messy.  let's do the
 *            messy diff.
 */
int  differ( int initial_rcol1, int initial_rcol2, int bottom )
{
int  rcol1;             /* the real cursor column on diff window 1 */
int  rcol2;             /* the real cursor column on diff window 2 */
int  r1;                /* the real text under rcol1 - needed for tabs */
int  r2;                /* the real text under rcol2 - needed for tabs */
char c1;                /* character under r1 */
char c2;                /* character under r2 */
int  leading1;          /* adjustment for leading space in window 1 */
int  leading2;          /* adjustment for leading space in window 2 */
int  len;               /* line length variable */
text_ptr diff1;         /* scratch text ptr in window 1 */
text_ptr diff2;         /* scratch text ptr in window 2 */
char line_buff[(MAX_COLS+1)*2];  /* buffer for char and attribute  */

   /*
    * initialize the text pointers and the initial column.  skip any
    *  initial blank lines.
    */
   if (diff.blank_lines) {
      while (diff.d1 != NULL  &&  is_line_blank( diff.d1 ))
         diff.d1 = find_next( diff.d1 );
      while (diff.d2 != NULL  &&  is_line_blank( diff.d2 ))
         diff.d2 = find_next( diff.d2 );
   }

   /*
    * use diff1 and diff2 to walk thru the text.
    * use diff.d1 and diff.d2 to walk to the next line.
    */
   diff1 = diff.d1;
   diff2 = diff.d2;
   if (diff1 != NULL  &&  diff2 != NULL) {
      rcol1 = initial_rcol1;
      rcol2 = initial_rcol2;

      /*
       * if cursors are past EOL, move them back to EOL.
       */
      len = find_end( diff1 );
      if (rcol1 > len)
         rcol1 = len;
      len = find_end( diff2 );
      if (rcol2 > len)
         rcol2 = len;

      /*
       * if skip leading space, make sure our cursors start on first non-space.
       */
      if (diff.leading) {
         leading1 = skip_leading_space( diff1 );
         leading2 = skip_leading_space( diff2 );
         if (mode.inflate_tabs) {
            leading1 = detab_adjust_rcol( diff1, leading1 );
            leading2 = detab_adjust_rcol( diff2, leading2 );
         }
         if (rcol1 < leading1)
            rcol1 = leading1;
         if (rcol2 < leading2)
            rcol2 = leading2;
      }

      /*
       * we now have a valid rcol for the diff start, we may need to adjust
       *   for tabs, though.
       */
      r1 = rcol1;
      r2 = rcol2;
      if (mode.inflate_tabs) {
         r1 = entab_adjust_rcol( diff1, rcol1 );
         r2 = entab_adjust_rcol( diff2, rcol2 );
      } else {
         diff1 += r1;
         diff2 += r2;
      }
      s_output( diff_message, g_display.mode_line, 67, g_display.diag_color );
      while (*diff1 != CONTROL_Z  &&  *diff2 != CONTROL_Z  && !g_status.control_break) {

         /*
          * if tabs are out, let's handle them differently.
          */
         if (mode.inflate_tabs) {
            c1 = *(diff1 + r1);
            c2 = *(diff2 + r2);

            /*
             *  tabs == space
             */
            if (c1 == '\t')
               c1 = ' ';
            if (c2 == '\t')
               c2 = ' ';
            if (diff.all_space) {
               while (c1 == ' ') {
                  ++rcol1;
                  r1 = entab_adjust_rcol( diff1, rcol1 );
                  c1 = *(diff1 + r1);
                  if (c1 == '\t')
                     c1 = ' ';
               }
               while (c2 == ' ') {
                  ++rcol2;
                  r2 = entab_adjust_rcol( diff1, rcol2 );
                  c2 = *(diff1 + r2);
                  if (c2 == '\t')
                     c2 = ' ';
               }
            }

            if (diff.ignore_eol) {
               if (c1 == '\n') {
                  diff1 = diff.d1 = skip_eol( diff.d1, &r1, &rcol1 );
                  c1 = *(diff1 + r1);
                  if (c1 == '\t')
                     c1 = ' ';
               }
               if (c2 == '\n') {
                  diff2 = diff.d2 = skip_eol( diff.d2, &r2, &rcol2 );
                  c1 = *(diff2 + r2);
                  if (c2 == '\t')
                     c2 = ' ';
               }
            }
            if (mode.search_case == IGNORE) {
               c1 = (char)tolower( c1 );
               c2 = (char)tolower( c2 );
            }
            while (c1 != '\n' && c2 != '\n' && c1 != CONTROL_Z && c2 != CONTROL_Z) {
               if (c1 == c2) {
                  if (diff.all_space) {
                     do {
                        ++rcol1;
                        r1 = entab_adjust_rcol( diff1, rcol1 );
                        c1 = *(diff1 + r1);
                        if (c1 == '\t')
                           c1 = ' ';
                     } while (c1 == ' ');
                     do {
                        ++rcol2;
                        r2 = entab_adjust_rcol( diff1, rcol2 );
                        c2 = *(diff1 + r2);
                        if (c2 == '\t')
                           c2 = ' ';
                     } while (c2 == ' ');
                  } else {
                     ++rcol1;
                     ++rcol2;
                     r1 = entab_adjust_rcol( diff1, rcol1 );
                     r2 = entab_adjust_rcol( diff2, rcol2 );
                     c1 = *(diff1 + r1);
                     c2 = *(diff2 + r2);
                     if (c1 == '\t')
                        c1 = ' ';
                     if (c2 == '\t')
                        c2 = ' ';
                  }
                  if (diff.ignore_eol) {
                     if (c1 == '\n') {
                        diff1 = diff.d1 = skip_eol( diff.d1, &r1, &rcol1 );
                        c1 = *(diff1 + r1);
                        if (c1 == '\t')
                           c1 = ' ';
                     }
                     if (c2 == '\n') {
                        diff2 = diff.d2 = skip_eol( diff.d2, &r2, &rcol2 );
                        c1 = *(diff2 + r2);
                        if (c2 == '\t')
                           c2 = ' ';
                     }
                  }
                  if (mode.search_case == IGNORE) {
                     c1 = (char)tolower( c1 );
                     c2 = (char)tolower( c2 );
                  }
               } else {

                  /*
                   * when we show the diff, we need to adjust the rcol
                   *   that comes out of find_adjust.  find_adjust does
                   *   not adjust rcol for tabs.
                   */
                  update_line( diff.w1 );
                  find_adjust( diff.w1, diff1 );
                  check_virtual_col( diff.w1, rcol1, rcol1 );
                  show_diff_window( diff.w1 );
                  update_line( diff.w2 );
                  find_adjust( diff.w2, diff2 );
                  check_virtual_col( diff.w2, rcol2, rcol2 );
                  show_diff_window( diff.w2 );
                  s_output( diff_blank, g_display.mode_line, 67,
                            g_display.mode_color );
                  return( OK );
               }
            }
         } else {

            /*
             * handle regular text here
             */
            c1 = *diff1;
            c2 = *diff2;
            if (diff.ignore_eol) {
               if (c1 == '\n') {
                  diff.d1 = skip_eol( diff.d1, &r1, &rcol1 );
                  diff1 = diff.d1 + r1;
                  c1 = *diff1;
               }
               if (c2 == '\n') {
                  diff.d2 = skip_eol( diff.d2, &r2, &rcol2 );
                  diff2 = diff.d2 + r2;
                  c2 = *diff2;
               }
            }
            if (diff.all_space) {
               if (c1 == ' ')
                  while ((c1 = *++diff1) == ' ');
               if (c2 == ' ')
                  while ((c2 = *++diff2) == ' ');
            }
            if (mode.search_case == IGNORE) {
               c1 = (char)tolower( c1 );
               c2 = (char)tolower( c2 );
            }
            while (c1 != '\n' && c2 != '\n' && c1 != CONTROL_Z && c2 != CONTROL_Z) {
               if (c1 == c2) {
                  if (diff.all_space) {
                     while ((c1 = *++diff1) == ' ');
                     while ((c2 = *++diff2) == ' ');
                  } else {
                     c1 = *++diff1;
                     c2 = *++diff2;
                  }
                  if (diff.ignore_eol) {
                     if (c1 == '\n') {
                        diff.d1 = skip_eol( diff.d1, &r1, &rcol1 );
                        diff1 = diff.d1 + r1;
                        c1 = *diff1;
                     }
                     if (c2 == '\n') {
                        diff.d2 = skip_eol( diff.d2, &r2, &rcol2 );
                        diff2 = diff.d2 + r2;
                        c2 = *diff2;
                     }
                  }
                  if (mode.search_case == IGNORE) {
                     c1 = (char)tolower( c1 );
                     c2 = (char)tolower( c2 );
                  }
               } else {
                  update_line( diff.w1 );
                  find_adjust( diff.w1, diff1 );
                  show_diff_window( diff.w1 );
                  update_line( diff.w2 );
                  find_adjust( diff.w2, diff2 );
                  show_diff_window( diff.w2 );
                  s_output( diff_blank, g_display.mode_line, 67,
                            g_display.mode_color );
                  return( OK );
               }
            }
         }

         /*
          * if we haven't come to the end of a file buffer, check the last
          *   characters.  because we break out of the while loop above at
          *   EOL or EOF, c1 may be '\n' and c2 may be something else, or
          *   vice versa.
          */
         if (c1 != CONTROL_Z && c2 != CONTROL_Z) {
            if (c1 != c2) {
               update_line( diff.w1 );
               find_adjust( diff.w1, diff1 );
               show_diff_window( diff.w1 );
               update_line( diff.w2 );
               find_adjust( diff.w2, diff2 );
               show_diff_window( diff.w2 );
               s_output( diff_blank, g_display.mode_line, 67,
                         g_display.mode_color );
               return( OK );
            }
         }
         if (c1 == c2  &&  c1 != CONTROL_Z) {
            diff1 = diff.d1 = skip_eol( diff.d1, &r1, &rcol1 );
            diff2 = diff.d2 = skip_eol( diff.d2, &r2, &rcol2 );
            if (!mode.inflate_tabs) {
               diff1 += r1;
               diff2 += r2;
            }
         }
      }
      save_screen_line( 0, bottom, line_buff );
      set_prompt( diff_prompt4, bottom );
      getkey( );
      restore_screen_line( 0, bottom, line_buff );
      s_output( diff_blank, g_display.mode_line, 67, g_display.mode_color );
   }
   return( ERROR );
}


/*
 * Name:    skip_leading_space
 * Purpose: put the diff on the first non-blank character
 * Date:    October 31, 1992
 * Passed:  s:  the string to search
 * Returns: the first non-blank column
 */
int  skip_leading_space( text_ptr s )
{
register int count = 0;

   s = cpf( s );
   if (mode.inflate_tabs) {
      while (*s == ' ' || *s == '\t') {
         ++count;
         ++s;
      }
   } else {
      while (*s == ' ') {
        ++count;
        ++s;
      }
   }
   if (*s == '\n' || *s == CONTROL_Z)
      count = 0;
   return( count );
}


/*
 * Name:    skip_eol
 * Purpose: move the diff to the next line
 * Date:    October 31, 1992
 * Passed:  s:  the string to search
 * Returns: the first non-blank column
 */
text_ptr skip_eol( text_ptr d, int *r, int *rcol )
{
int leading;
text_ptr q;

   *r = *rcol = 0;
   q = d;
   d = find_next( d );
   if (diff.blank_lines) {
      while (d != NULL  &&  is_line_blank( d )) {
         q = d;
         d = find_next( d );
      }
   }
   if (d != NULL) {
      if (diff.leading) {
         leading = skip_leading_space( d );
         if (mode.inflate_tabs)
            leading = detab_adjust_rcol( d, leading );
         *rcol = leading;
      } else
         *rcol = 0;
      *r = *rcol;
      if (mode.inflate_tabs)
         *r = entab_adjust_rcol( d, *rcol );
   } else
      d = q;
   return( d );
}


/*
 * Name:    show_diff_window
 * Purpose: update the contents of a diff window
 * Date:    October 31, 1992
 * Passed:  win:  pointer to window
 */
void show_diff_window( WINDOW *win )
{
   if (win->file_info->dirty == LOCAL)
      display_current_window( win );
   else
      show_curl_line( win );
   show_line_col( win );
   make_ruler( win );
   show_ruler( win );
   show_ruler_pointer( win );
   win->file_info->dirty = FALSE;
}


/*
 * Name:    verify_number
 * Purpose: given a window number, verify the number
 * Date:    October 31, 1992
 * Passed:  temp:  string that contains number
 *          num:   successfully converted number.
 */
int  verify_number( char *temp, int *num )
{
register file_infos *fp;

   /*
    * see if string has a number.  if string does have a number, convert it.
    */
   if (*temp == '\0' || !isdigit( *temp ))
      return( ERROR );
   *num = 0;
   while (isdigit( *temp ))
      *num = *num * 10 + *temp++ - '0';

   /*
    * now that we have a window number, see if any files have that number
    */
   for (fp=g_status.file_list; fp != NULL; fp=fp->next) {
      if (fp->file_no == *num)
         return( OK );
   }
   return( ERROR );
}



/*
 * Name:    verify_letter
 * Purpose: given a window letter, verify the letter
 * Date:    October 31, 1992
 * Passed:  temp:  string that contains number
 *          num:   successfully converted number.
 */
int  verify_letter( char *temp, int *let, WINDOW **win )
{
register file_infos *fp;
register WINDOW *wp;
int num;

   if (verify_number( temp, &num ) == OK) {
      while (isdigit( *temp ))
         temp++;
      if (*temp == '\0' || !isalpha( *temp ))
         return( ERROR );
      *let = (int)tolower( *temp );
      for (fp=g_status.file_list; fp != NULL; fp=fp->next) {
         if (fp->file_no == num)
            break;
      }
      for (wp=g_status.window_list; wp != NULL; wp=wp->next) {
         if (wp->file_info == fp  &&  wp->letter == *let  &&  wp->visible) {
            *win = wp;
            return( OK );
         }
      }
   }
   return( ERROR );
}
