/*******************  start of original comments  ********************/
/*
 * Written by Douglas Thomson (1989/1990)
 *
 * This source code is released into the public domain.
 */

/*
 * Name:    dte - Doug's Text Editor program - find/replace module
 * Purpose: This file contains the functions relating to finding text
 *           and replacing text.
 *          It also contains the code for moving the cursor to various
 *           other positions in the file.
 * File:    findrep.c
 * Author:  Douglas Thomson
 * System:  this file is intended to be system-independent
 * Date:    October 1, 1989
 */
/*********************  end of original comments   ********************/

/*
 * The search and replace routines have been EXTENSIVELY rewritten.  The
 * "brute force" search algorithm has been replaced by the Boyer-Moore
 * string search algorithm.  This search algorithm really speeds up search
 * and replace operations.
 *
 * In 2.1, the find functions were slightly modified.  Dead space is left
 * at the beginning and ending of the main text buffer.  The dead space is
 * used by sentinels in the Boyer-Moore search algorithm.  Sentinels are
 * used to keep the search from going past either end of the main text
 * buffer.  If nothing is used to stop the B-M search, it will wander out of
 * the main text buffer and into video memory, ROM, system memory, etc...
 * Although the old search algorithm was fast enough, using a sentinel sqeezes
 * a little more speed from the search functions.  Searching for patterns in
 * large files is noticably faster.  Part of this speed increase is due to
 * improving function find_adjust( ), which adjusts the screen after a search.
 *
 * If I am not mistook, seems like Boyer developed the array and Moore
 * developed the search.  For those interested, the algorithm is published in:
 *
 *    Robert S. Boyer and J Strother Moore, "A fast string searching
 *      algorithm."  _Communications of the ACM_ 20 (No. 10): 762-772, 1977.
 *
 *  See also:
 *
 *    Zvi Galil, "On Improving the Worst Case Running Time of the Boyer-
 *      Moore String Matching Algorithm."  _Communications of the ACM_
 *      22 (No. 9): 505-508, 1979.
 *
 *    R. Nigel Horspool, "Practical fast searching in strings." _Software-
 *      Practice and Experience._  10 (No. 3): 501-506, 1980.
 *
 *    Alberto Apostolico and Raffaele Giancarlo, "The Boyer-Moore-Galil
 *      String Searching Strategies Revisited."  _SIAM J. Comput._ 15 (No. 1):
 *      98-105, 1986
 *
 *    Timo Raita, "Tuning the Boyer-Moore-Horspool String Searching Algorithm."
 *      _Software-Practice and Experience_ 22 (No. 10): 879-884, October 1992.
 *
 *  See also:
 *
 *    Andrew Hume and Daniel Sunday, "Fast String Searching."  _Software-
 *      Practice and Experience_ 21 (No. 11): 1221-1248, November 1991.
 *
 *  See also:
 *
 *    Ricardo Baeza-Yates and Gaston H. Gonnet, "A New Approach to Text
 *      Searching."  _Communications of the ACM_ 35 (No. 10): 74-91,
 *      October 1992.
 *
 *    Sun Wu and Udi Manber, "Fast Text Searching Allowing Errors."
 *      _Communications of the ACM_ 35 (No. 10): 83-91, October 1992.
 *
 *                          Boyer-Moore in TDE, 2.10
 *
 * Hume and Sunday demonstrated in their paper that using a simple shift
 * function after a mismatch gives one of the fastest search times with the
 * Boyer-Moore algorithm.  When searching normal text for small patterns
 * (patterns less than 30 characters or so), Hume and Sunday found that the
 * faster the algorithm can get back into the fast skip loop with the
 * largest shift value then the faster the search.  Some algorithms can
 * generate a large shift value, but they can't get back into the skip loop
 * very fast.  Hume and Sunday give a simple method for computing a shift
 * constant that, more often than not, is equal to the length of the search
 * pattern.  They refer to the constant as mini-delta2 or md2.  From the
 * end of the string, md2 is the first leftmost occurrence of the rightmost
 * character.  Hume and Sunday also found that using a simple string
 * compair as the match function gave better performance than using the C
 * library memcmp( ) function.  By the time the memcmp( ) function initializes,
 * does its thing, and returns from a call, a simple string compare loop
 * could have already compared the strings.  See the Hume and Sunday paper
 * for an excellent discussion on fast string searching.  Incidentally,
 * Hume and Sunday describe an even faster, tuned Boyer-Moore search function.
 *
 * The Boyer-Moore search algorithm in TDE was rearranged so that it is now
 * almost identical to the original version Boyer-Moore described in CACM.
 * The simple shift function in TDE was replaced by the mini-delta2 shift
 * function in the Hume and Sunday paper.
 *
 * I am not very fond of WordStar/TURBO x style search and replace prompting.
 * Once the search pattern has been defined, one only needs to press a key
 * to search forwards or backwards.  The forward or backward search key may
 * be pressed at any time in any file once the pattern has been entered.  Also,
 * the search case may be toggled at any time in any file once the pattern has
 * has been entered.
 *
 * New editor name:  TDE, the Thomson-Davis Editor.
 * Author:           Frank Davis
 * Date:             June 5, 1991, version 1.0
 * Date:             July 29, 1991, version 1.1
 * Date:             October 5, 1991, version 1.2
 * Date:             January 20, 1992, version 1.3
 * Date:             February 17, 1992, version 1.4
 * Date:             April 1, 1992, version 1.5
 * Date:             June 5, 1992, version 2.0
 * Date:             October 31, 1992, version 2.1
 *
 * This modification of Douglas Thomson's code is released into the
 * public domain, Frank Davis.   You may distribute it freely.
 */

#include "tdestr.h"
#include "common.h"
#include "tdefunc.h"
#include "define.h"


/*
 * Name:    get_replacement_flags
 * Purpose: To input find and replace flags.
 * Date:    June 5, 1991
 * Passed:  line:  prompt line
 * Returns: OK if flags were entered, ERROR if user wanted to abort
 */
int  get_replacement_flags( int line )
{
char line_buff[(MAX_COLS+1)*2]; /* buffer for char and attribute  */
register int c;
int rc, func;

   save_screen_line( 0, line, line_buff );
   eol_clear( 0, line, g_display.text_color );
   /*
    * options: prompt or no prompt (p/n)?
    */
   s_output( find1, line, 0, g_display.message_color );
   xygoto( strlen( find1 )+2, line );
   do {
      c = getkey( );
      func = getfunc( c );
   } while (c != 'P'  &&  c != 'p'  &&  c != 'N'  &&  c != 'n'  &&
            c != RTURN  &&  c != ESC  &&  func != AbortCommand);
   restore_screen_line( 0, line, line_buff );
   switch (c) {
      case 'P' :
      case 'p' :
      case RTURN :
         g_status.replace_flag = PROMPT;
         rc = OK;
         break;
      case 'N' :
      case 'n' :
         g_status.replace_flag = NOPROMPT;
         rc = OK;
         break;
      default :
         rc = ERROR;
   }
   bm.search_defined = rc;
   return( rc );
}


/*
 * Name:    ask_replace
 * Purpose: Ask user if he wants to (r)place, (s)kip, or (e)xit.
 * Date:    June 5, 1991
 * Returns: TRUE if user wants to replace, ERROR otherwise.
 * Passed:  window:   pointer to current window
 *          finished: TRUE if user pressed ESC or (e)xit.
 */
int  ask_replace( WINDOW *window, int *finished )
{
char line_buff[(MAX_COLS+1)*2]; /* buffer for char and attribute  */
register int c;
int rc, prompt_line, func;

   prompt_line = window->cline - 1;
   c = 39 - (strlen( find2 ) >> 1);
   save_screen_line( 0, prompt_line, line_buff );
   /*
    * replace skip exit (r/s/e)?
    */
   s_output( find2, prompt_line, c, g_display.message_color );
   do {
      c = getkey( );
      func = getfunc( c );
   } while (c != 'R' && c != 'r' && c != 'S' && c != 's' && c != 'E' && c != 'e'
          && c != ESC  &&  func != AbortCommand);
   restore_screen_line( 0, prompt_line, line_buff );
   switch (c) {
      case 'R' :
      case 'r' :
         rc = OK;
         break;
      case 'E' :
      case 'e' :
         *finished = TRUE;
      case 'S' :
      case 's' :
         rc = ERROR;
         break;
      default :
         *finished = TRUE;
         rc = ERROR;
         break;
   }
   return( rc );
}


/*
 * Name:    ask_wrap_replace
 * Purpose: After a wrap, ask user if he wants to (q)uit or (c)ontine.
 * Date:    June 5, 1991
 * Returns: TRUE if user wants to continue, ERROR otherwise.
 * Passed:  window:   pointer to current window
 *          finished: TRUE if user pressed ESC or (q)uit.
 */
int  ask_wrap_replace( WINDOW *window, int *finished )
{
char line_buff[(MAX_COLS+1)*2]; /* buffer for char and attribute  */
register int c;
int rc, prompt_line, func;

   prompt_line = window->bottom_line;
   save_screen_line( 0, prompt_line, line_buff );
   /*
    * search has wrapped. continue or quit (c/q)?
    */
   set_prompt( find3, prompt_line );
   do {
      c = getkey( );
      func = getfunc( c );
   } while (c != 'Q'  &&  c != 'q'  &&  c != 'C'  &&  c != 'c' &&
          c != ESC  &&  func != AbortCommand);
   restore_screen_line( 0, prompt_line, line_buff );
   switch (c) {
      case 'C' :
      case 'c' :
         rc = OK;
         break;
      case 'Q' :
      case 'q' :
      default :
         *finished = TRUE;
         rc = ERROR;
         break;
   }
   return( rc );
}


/*
 * Name:    do_replace
 * Purpose: To replace text once match has been found.
 * Date:    June 5, 1991
 * Passed:  window:     pointer to current window
 *          start:      location of start of matched text
 *          direction:  direction of search
 */
void do_replace( WINDOW *window, text_ptr start, int direction )
{
int old_len;             /* length of original text */
int new_len;             /* length of replacement text */
register int net_change;
text_ptr source;         /* source of block move */
text_ptr dest;           /* destination of block move */
long number;             /* number of characters moved */

   old_len = strlen( g_status.pattern );
   new_len = strlen( g_status.subst );

   /*
    * move the text to either make room for the extra replacement text
    *  or to close up the gap left
    */
   start = cpf( start );
   source = start + old_len;
   dest = start + new_len;
   number = ptoul( g_status.end_mem ) - ptoul( source );
   hw_move( dest, source, number );

   /*
    * insert the replacement text
    */
   for (dest=start, source=g_status.subst; *source;)
      *dest++ = *source++;

   net_change = new_len - old_len;
   adjust_start_end( window->file_info, net_change );
   if (direction == FORWARD && net_change > 0)
      window->rcol += net_change;
   addorsub_all_cursors( window, net_change );
   g_status.end_mem = addltop( (long)net_change, g_status.end_mem );
   if (mode.inflate_tabs) {
      old_len = g_status.command;
      g_status.command = WordWrap;
      g_status.copied = FALSE;
      copy_line( window->cursor, window->bottom_line );
      detab_linebuff( );
      entab_linebuff( );
      un_copy_line( window->cursor, window, TRUE );
      g_status.command = old_len;
   }
   show_avail_mem( );
}


/*
 * Name:    find_string
 * Purpose: To set up and perform a find operation.
 * Date:    June 5, 1991
 * Passed:  window:  pointer to current window
 * Notes:   Keep the search string and boyer-moore stuff until changed.
 */
int  find_string( WINDOW *window )
{
int  direction;
int  new_string;
char pattern[MAX_COLS];  /* text to be found */
text_ptr pattern_location;
register WINDOW *win;  /* put window pointer in a register */
int  rc;
int  old_rcol;

   switch (g_status.command) {
      case FindForward :
         direction = FORWARD;
         new_string = TRUE;
         break;
      case FindBackward :
         direction = BACKWARD;
         new_string = TRUE;
         break;
      case RepeatFindForward1 :
      case RepeatFindForward2 :
         direction = FORWARD;
         new_string = FALSE;
         break;
      case RepeatFindBackward1 :
      case RepeatFindBackward2 :
         direction = BACKWARD;
         new_string = FALSE;
         break;
   }
   win = window;
   entab_linebuff( );
   un_copy_line( win->cursor, win, TRUE );
   /*
    * get search text, using previous as default
    */
   if (new_string == TRUE) {
      strcpy( pattern, bm.pattern );
      /*
       * string to find:
       */
      if (get_name( find4, win->bottom_line, pattern,
                    g_display.message_color ) != OK)
         return( ERROR );
      bm.search_defined = OK;
      strcpy( bm.pattern, pattern );

      build_boyer_array( );
   }

   rc = OK;
   if (bm.search_defined == OK) {
      if (mode.inflate_tabs) {
         old_rcol = win->rcol;
         win->rcol = entab_adjust_rcol( win->cursor, win->rcol );
      }
      update_line( win );
      show_search_message( SEARCHING, g_display.diag_color );
      if (direction == FORWARD) {
         if ((pattern_location = forward_boyer_moore_search( win )) != NULL) {
            if (g_status.wrapped && g_status.macro_executing)
               rc = ask_wrap_replace( win, &new_string );
            if (rc == OK)
               find_adjust( win, pattern_location );
         }
      } else {
         if ((pattern_location = backward_boyer_moore_search( win )) != NULL) {
            if (g_status.wrapped && g_status.macro_executing)
               rc = ask_wrap_replace( win, &new_string );
            if (rc == OK)
               find_adjust( win, pattern_location );
         }
      }
      if (g_status.wrapped)
         show_search_message( WRAPPED, g_display.diag_color );
      else
         show_search_message( CLR_SEARCH, g_display.mode_color );
      if (pattern_location == NULL) {
         /*
          * string not found
          */
         if (mode.inflate_tabs)
            win->rcol = old_rcol;
         combine_strings( pattern, find5a, bm.pattern, find5b );
         error( WARNING, win->bottom_line, pattern );
         rc = ERROR;
      }
      show_curl_line( win );
      make_ruler( win );
      show_ruler( win );
   } else {
      /*
       * find pattern not defined
       */
      error( WARNING, win->bottom_line, find6 );
      rc = ERROR;
   }
   return( rc );
}


/*
 * Name:    build_boyer_array
 * Purpose: To set up the boyer array for forward and backward searches.
 * Date:    June 5, 1991
 * Notes:   Set up one array for forward searches and another for backward
 *          searches.  If user decides to ignore case then fill in array
 *          with reverse case characters so both upper and lower case
 *          characters are defined.
 */
void build_boyer_array( void )
{
   /*
    * set up for forward search
    */
   if (g_status.command == DefineSearchAndSeize  ||
                           g_status.command == RepeatSearchAndSeize) {
      memcpy( bm.pattern, sas_bm.pattern, strlen( sas_bm.pattern ) + 1 );
      bm.search_defined = OK;
   }

   if (bm.search_defined == OK) {
      build_forward_skip( &bm );
      bm.forward_md2 = calculate_forward_md2( bm.pattern, bm.pattern_length );

      /*
       * set up for backward search
       */
      build_backward_skip( &bm );
      bm.backward_md2 = calculate_backward_md2( bm.pattern, bm.pattern_length );
   }


   /*
    * build an array for search and seize.
    */
   if (sas_bm.search_defined == OK) {
      build_forward_skip( &sas_bm );
      sas_bm.forward_md2 = calculate_forward_md2( sas_bm.pattern,
                                                  sas_bm.pattern_length );

      /*
       * set up for backward search
       */
      build_backward_skip( &sas_bm );
      sas_bm.backward_md2 = calculate_backward_md2( sas_bm.pattern,
                                                    sas_bm.pattern_length );
   }
}


/*
 * Name:    build_forward_skip
 * Purpose: build Boyer-Moore forward skip array
 * Date:    October 31, 1992
 * Notes:   build standard Boyer-Moore forward skip array.
 */
void build_forward_skip( boyer_moore_type *bmp )
{
register unsigned char *p;
register int i;

   i = bmp->pattern_length = strlen( bmp->pattern );

   /*
    * set skip index of all characters to length of string
    */
   memset( bmp->skip_forward, i, 256 );
   i--;

   /*
    * for each character in string, set the skip index
    */
   for (p=bmp->pattern; i>=0; i--, p++) {
      bmp->skip_forward[*p] = (char)i;
      if (mode.search_case == IGNORE) {
         if (*p >= 'A' && *p <= 'Z')
            bmp->skip_forward[*p+32] = (char)i;
         else if (*p >= 'a' && *p <= 'z')
            bmp->skip_forward[*p-32] = (char)i;
      }
   }
}


/*
 * Name:    build_backward_skip
 * Purpose: build Boyer-Moore backward skip array
 * Date:    October 31, 1992
 * Notes:   build standard Boyer-Moore backward skip array.
 */
void build_backward_skip( boyer_moore_type *bmp )
{
register unsigned char *p;
register int i;

   i = -bmp->pattern_length;
   memset( bmp->skip_backward, i, 256 );
   i++;
   for (p=bmp->pattern + bmp->pattern_length - 1; i<=0; i++, p--) {
      bmp->skip_backward[*p] = (char)i;
      if (mode.search_case == IGNORE) {
         if (*p >= 'A' && *p <= 'Z')
            bmp->skip_backward[*p+32] = (char)i;
         else if (*p >= 'a' && *p <= 'z')
            bmp->skip_backward[*p-32] = (char)i;
      }
   }
}


/*
 * Name:    calculate_forward_md2
 * Purpose: Calculate mini delta2 for forward searches
 * Date:    October 31, 1992
 * Notes:   Hume and Sunday (see above) demonstrate in their paper that
 *            using a simple shift function on mismatches with BM
 *            gives one of the fastest run times for general text searching.
 *          mini delta2 is, from the end of the string, the leftmost
 *            occurrence of the rightmost character.  mini delta2 is 1 in
 *            the worst case.  in previous versions of TDE, the shift function
 *            was hard-coded as 1 -- the worst case.  typically, mini delta2
 *            will be the length of the search pattern.
 */
int  calculate_forward_md2( char *p, int len )
{
int last_c;
register int i;
register int md2;

   md2 = 1;
   i = len - 1;
   last_c = p[i];
   if (mode.search_case == IGNORE) {
      last_c = tolower( last_c );
      for (i--; i >= 0  &&  last_c != tolower( p[i] ); i--)
         ++md2;
   } else {
      for (i--; i >= 0  &&  last_c != p[i]; i--)
         ++md2;
   }
   return( md2 );
}


/*
 * Name:    calculate_backward_md2
 * Purpose: Calculate mini delta2 for backward searches
 * Date:    October 31, 1992
 * Notes:   the backward mini delta2 is, from the start of the string, the
 *            first rightmost occurence of the leftmost character.  in the
 *            worst case, mini delta2 is -1.  typically, mini delta2 is the
 *            negative length of the search pattern.
 */
int  calculate_backward_md2( char *p, int len )
{
int first_c;
register int i;
register int md2;

   md2 = -1;
   i = 1;
   first_c = *p;
   if (mode.search_case == IGNORE) {
      first_c = tolower( first_c );
      for (; i < len  &&  first_c != tolower( p[i] ); i++)
         --md2;
   } else {
      for (; i < len  &&  first_c != p[i]; i++)
         --md2;
   }
   return( md2 );
}


/*
 * Name:    forward_boyer_moore_search
 * Purpose: search forward for pattern using boyer array
 * Passed:  window:  pointer to current window
 * Returns: position in file if found or NULL if not found
 * Date:    June 5, 1991
 * Notes:   Start searching from cursor position to end of file.  If we hit
 *          end of file without a match, start searching from the beginning
 *          of file to cursor position.  (do wrapped searches)
 */
text_ptr forward_boyer_moore_search( WINDOW *window )
{
int i;
register int len;
text_ptr s, start;
text_ptr end;
register WINDOW *win;  /* put window pointer in a register */
char sentinel[MAX_COLS];

   /*
    * if cursor is beyond end of line then start at end of line
    */
   win  = window;
   start = cpf( win->cursor );
   i = win->rcol + 1;
   len = linelen( start );
   if (i > len) {
      i = len;
      len = 0;
   } else
      len = i + bm.pattern_length - 2 >= len ? len - i : bm.pattern_length - 2;

   /*
    * make start of search 1 character greater than current position so
    * we don't repeatedly find the pattern at current position.
    */
   start += i;

   end = cpf( win->file_info->end_text );

   /*
    * put a sentinel at the end of the file.
    */
   _fmemcpy( sentinel, end, bm.pattern_length );
   _fmemcpy( end, bm.pattern, bm.pattern_length );

   s = search_forward( start );

   /*
    * remove the sentinel and get back the original text
    */
   _fmemcpy( end, sentinel, bm.pattern_length );

   if (ptoul( s ) == ptoul( end )) {

      /*
       * haven't found pattern yet - now search from beginning of file
       */
      g_status.wrapped = TRUE;

      /*
       * for the wrapped search, put the sentinel at start
       */
      start += len;
      _fmemcpy( sentinel, start, bm.pattern_length );
      _fmemcpy( start, bm.pattern, bm.pattern_length );
      s = search_forward( cpf( win->file_info->start_text ) );
      _fmemcpy( start, sentinel, bm.pattern_length );
      if ( ptoul( s ) == ptoul( start ))
         s = NULL;
   }
   return( s );
}


/*
 * Name:    search_forward
 * Purpose: search forward for pattern using boyer array
 * Passed:  s:              text_ptr for search start
 *          search_length:  number of characters to search
 * Returns: position in file if found or NULL if not found
 * Date:    January 8, 1992
 * Notes:   start searching from cursor position to end of file.
 *            a sentinel is added at the end of the text buffer to keep the
 *            search from go past the EOF.  the search will either find the
 *            sentinel or the pattern in the text buffer.
 *          mini delta2 is the leftmost occurrence of the rightmost character.
 */
text_ptr search_forward( text_ptr s )
{
register int i;
register int j;
text_ptr p;
text_ptr q;
int mini_delta2;
int len;
char *skip;
boyer_moore_type *bmp;

   if (g_status.command == DefineSearchAndSeize ||
       g_status.command == RepeatSearchAndSeize)
      bmp = &sas_bm;
   else
      bmp = &bm;

   mini_delta2 = bmp->forward_md2;
   len  = bmp->pattern_length;
   skip = bmp->skip_forward;
   p    = bmp->pattern;

   j = i = len - 1;
   if (mode.search_case == MATCH) {
      for (;;) {

         /*
          * Boyer-Moore fast skip loop.  we need to check our text pointer as
          *   we move thru the text.
          */
         for (s+=i, j+=i; i = skip[(unsigned char)*s]; s+=i, j+=i) {

            /*
             * every 10,000 characters, check the text pointer
             */
            if (j > 10000) {
               j = 0;
               s = cpf( s );
            }
         }

         /*
          * i == 0, possible match.  Boyer-Moore slow match loop.
          */
         q = cpf( addltop( 1 - len, s ) );
         for (i=0; i<len; i++) {
            if (q[i] != p[i])
              goto fast_match;
         }
         return( q );
fast_match:
         i = mini_delta2;
      }
   } else {
      for (;;) {
         for (s+=i, j+=i; i = skip[(unsigned char)*s]; s+=i, j+=i) {
            if (j > 10000) {
               j = 0;
               s = cpf( s );
            }
         }

         /*
          * i == 0, possible match.  Boyer-Moore slow match loop.
          */
         q = cpf( addltop( 1 - len, s ) );
         for (i=0; i<len; i++) {
            if (tolower( q[i] ) != tolower( p[i] ))
               goto fast_ignore;
         }
         return( q );
fast_ignore:
         i = mini_delta2;
      }
   }
}


/*
 * Name:    backward_boyer_moore_search
 * Purpose: search backward for pattern using boyer array
 * Passed:  window:  pointer to current window
 * Returns: position in file if found or NULL if not found
 * Date:    June 5, 1991
 * Notes:   Start searching from cursor position to beginning of file. If we
 *          hit beginning end of file without a match, start searching from the
 *          end of file to cursor position.  (do wrapped searches)
 */
text_ptr backward_boyer_moore_search( WINDOW *window )
{
int i;
int len;
text_ptr s;
text_ptr min;
register WINDOW *win;  /* put window pointer in a register */
char sentinel[MAX_COLS];

   win  = window;
   min = cpf( win->cursor );
   i = win->rcol - 1;
   i += bm.pattern_length - 1;
   len = linelen( min );
   if (i > len)
      i = len;
   if (len > win->rcol)
      len = win->rcol;


   /*
    * put a sentinel at the beginning of the file.
    */
   min = cpf( addltop( -bm.pattern_length, win->file_info->start_text ) );
   _fmemcpy( sentinel, min, bm.pattern_length );
   _fmemcpy( min, bm.pattern, bm.pattern_length );

   s = cpb( addltop( i, win->cursor ) );
   s = search_backward( s );

   /*
    * remove the sentinel and get back the original text
    */
   _fmemcpy( min, sentinel, bm.pattern_length );

   if (ptoul( s ) == ptoul( min )) {

      /*
       * haven't found pattern yet - now search from end of file
       */
      g_status.wrapped = TRUE;
      min = cpf( win->cursor ) + len;
      min = cpf( addltop( -bm.pattern_length+1, min ) );
      _fmemcpy( sentinel, min, bm.pattern_length );
      _fmemcpy( min, bm.pattern, bm.pattern_length );

      s = cpb( addltop( -1, win->file_info->end_text ) );
      s = search_backward( s );
      _fmemcpy( min, sentinel, bm.pattern_length );
      if ( ptoul( s ) == ptoul( min ))
         s = NULL;
   }
   return( s );
}


/*
 * Name:    search_backward
 * Purpose: search backward for pattern using boyer array
 * Passed:  s:              text_ptr for search start
 *          search_length:  number of characters to search
 * Returns: position in file if found else return NULL
 * Date:    January 8, 1992
 * Notes:   Start searching from cursor position to beginning of file.
 *            a sentinel was added at the beginning of the text buffer
 *            to keep the search from going behind the start of the file.
 *          mini delta2 is the rightmost occurrence of the leftmost character.
 */
text_ptr search_backward( text_ptr s )
{
register int i;
register int j;
text_ptr p;
text_ptr q;
int mini_delta2;
int len;

   mini_delta2 = bm.backward_md2;
   len  = bm.pattern_length;
   p = bm.pattern;

   j = i = -bm.pattern_length + 1;
   if (mode.search_case == MATCH) {
      for (;;) {

         /*
          * Boyer-Moore fast skip loop.  we need to check our text pointer as
          *   we move thru the text.
          */
         for (s+=i, j+=i; i = bm.skip_backward[(unsigned char)*s]; s+=i, j+=i) {

            /*
             * every 10,000 characters, check the text pointer
             */
            if (j < -10000) {
               j = 0;
               s = cpb( s );
            }
         }

         /*
          * i == 0, possible match.  Boyer-Moore slow match loop.
          */
         q = cpf( s );
         for (i=0; i<len; i++) {
            if (q[i] != p[i])
              goto fast_match;
         }
         return( q );
fast_match:
         i = mini_delta2;
      }
   } else {
      for (;;) {
         for (s+=i, j+=i; i = bm.skip_backward[(unsigned char)*s]; s+=i, j+=i) {
            if (j < -10000) {
               j = 0;
               s = cpb( s );
            }
         }

         /*
          * i == 0, possible match.  do BM slow loop.
          */
         q = cpf( s );
         for (i=0; i<len; i++) {
            if (tolower( q[i] ) != tolower( p[i] ))
               goto fast_ignore;
         }
         return( q );
fast_ignore:
         i = mini_delta2;
      }
   }
}


/*
 * Name:    show_search_message
 * Purpose: display search status
 * Date:    January 8, 1992
 * Passed:  i:     index into message array
 *          color: color to display message
 */
void show_search_message( int i, int color )
{
   /*
    *  0 = blank
    *  1 = wrapped...
    *  2 = searching
    */
   s_output( find7[i], g_display.mode_line, 67, color );
}


/*
 * Name:    find_adjust
 * Purpose: place cursor on screen given a position in file - default cline
 * Date:    June 5, 1991
 * Passed:  window:  pointer to current window
 *          found:  position anywhere in file
 * Notes:   found could be anywhere in file.  Find the start of line that
 *          found is on.  Determine if start of line is behind or ahead of
 *          current line.  Once that is done, it is easy to determine if found
 *          is on screen.  Lastly, current cursor position becomes start of
 *          found line - reposition and display.
 */
void find_adjust( WINDOW *window, text_ptr found )
{
int rcol;
int cmd;
long pattern_line, rline, test_line;
unsigned long top, bot;
unsigned long begin, end;
text_ptr p, q;
file_infos *file;
register WINDOW *win;  /* put window pointer in a register */

   win = window;
   file = win->file_info;
   /*
    * find start of line found is on.
    */
   found = cpb( found );
   if (*(found - 1) != '\n' && *(found - 1) != CONTROL_Z)
      p = find_prev( found );
   else
      p = found;

   /*
    * find real column found is on.
    */
   rcol = (int)(ptoul( found ) - ptoul( p ));
   if (mode.inflate_tabs)
      rcol = detab_adjust_rcol( p, rcol );
   rline = pattern_line = win->rline;
   q = win->cursor = cpf( win->cursor );
   begin = ptoul( win->file_info->start_text );
   end   = ptoul( win->file_info->end_text );

   /*
    * if p, start of found line, is greater than current line, see if
    * p is between current line and bottom line on screen.
    */
   if ((bot = ptoul( q ))  <  (top = ptoul( p ))) {

      if ((top - bot) / 2l <= (end - top) / 2l) {
         while (ptoul( q ) != top) {
            q = find_next( q );
            ++pattern_line;
         }
      } else {
         pattern_line = win->file_info->length;
         q = find_prev( cpb( win->file_info->end_text ) - 1 );
         while (ptoul( q ) != top) {
            q = find_prev( q );
            --pattern_line;
         }
      }

      /*
       * test_line is the number of lines between rline and found line.
       */
      test_line = pattern_line - rline;
      if ((long)win->cline + test_line <= (long)win->bottom_line)
         win->cline += test_line;
      else
         file->dirty = LOCAL;

   /*
    * if p, start of found line, is less than current line, see if
    * p is between current line and top line on screen.
    */
   } else if (bot > top) {

      if ((bot - top) / 2l  <=  (top - begin) / 2l) {
         q = cpb( q );
         while (ptoul( q ) != top) {
            q = find_prev( q );
            --pattern_line;
         }
      } else {
         pattern_line = 1l;
         q = cpf( win->file_info->start_text );
         while (ptoul( q ) != top) {
            q = find_next( q );
            ++pattern_line;
         }
      }


      test_line = rline - pattern_line;
      if ((long)win->cline - test_line > (long)(win->top_line+win->ruler-1))
         win->cline -= test_line;
      else
         file->dirty = LOCAL;
      if (pattern_line < (long)(win->cline - (win->top_line+win->ruler-1)))
         win->cline = (int)pattern_line + win->top_line+win->ruler - 1;
   }

   /*
    * cursor line becomes found line.  check if column is on screen.
    */
   win->cursor = p;
   win->rline = pattern_line;
   if (file->dirty == LOCAL && (win->cline == win->bottom_line ||
                                win->cline == win->top_line + win->ruler)) {
      cmd = g_status.command;
      if (cmd == RepeatFindForward1    ||  cmd == RepeatFindBackward1 ||
          cmd == ReplaceString         ||
          cmd == DefineDiff            ||  cmd == RepeatDiff          ||
          cmd == DefineSearchAndSeize  ||  cmd == RepeatSearchAndSeize) {
         g_status.command = CenterLine;
         center_window( win );
         g_status.command = cmd;
      }
   }
   check_virtual_col( win, rcol, rcol );
}


/*
 * Name:    replace_string
 * Purpose: To set up and perform a replace operation.
 * Date:    June 5, 1991
 * Passed:  window:  pointer to current window
 */
int  replace_string( WINDOW *window )
{
int direction;
char pattern[MAX_COLS];  /* the old and replacement text */
int net_change;
int sub_len;
int file_changed;
int finished;
int rc;
text_ptr pattern_location;
text_ptr replace_start;
unsigned long rs, pl;
WINDOW wp;
register WINDOW *win;  /* put window pointer in a register */

   win = window;
   direction = get_replace_direction( win );
   if (direction == ERROR)
      return( ERROR );

   entab_linebuff( );
   un_copy_line( win->cursor, win, TRUE );

   /*
    * get the search pattern, using the previous as the default
    */
   strcpy( pattern, g_status.pattern );
   /*
    * string to find
    */
   if (get_name( find9, win->bottom_line, pattern,
                 g_display.message_color ) != OK)
      return( ERROR );
   strcpy( g_status.pattern, pattern );

   /*
    * get the replacement text, using the previous as the default
    */
   strcpy( pattern, g_status.subst );
   if (get_name( find10, win->bottom_line, pattern,
                  g_display.message_color ) != OK)
      return( ERROR );
   strcpy( g_status.subst, pattern );
   sub_len = strlen( pattern );
   strcpy( bm.pattern, g_status.pattern );
   net_change = sub_len - strlen( g_status.pattern );

   /*
    * get the replace flags, Prompt or NoPrompt
    */
   if (get_replacement_flags( win->bottom_line ) != OK)
      return( ERROR );

   build_boyer_array( );
   dup_window_info( &wp, win );
   update_line( win );

   if (mode.inflate_tabs)
      win->rcol = entab_adjust_rcol( win->cursor, win->rcol );

   rc = OK;
   finished = FALSE;
   file_changed = FALSE;
   if (direction == FORWARD) {
      if ((replace_start = forward_boyer_moore_search( &wp )) != NULL &&
          !g_status.control_break) {
         rs = ptoul( replace_start );
         replace_and_display( &wp, replace_start, &finished, &file_changed,
                              direction );
      } else {
         /*
          * string not found
          */
         error( WARNING, win->bottom_line, find8 );
         finished = TRUE;
         rc = ERROR;
      }
      while (finished == FALSE) {
         update_line( &wp );
         if ((pattern_location = forward_boyer_moore_search( &wp )) != NULL &&
             !g_status.control_break) {
            pl = ptoul( pattern_location );
            if (rs == pl)
               finished = TRUE;
            else {
               rc = replace_and_display( &wp, pattern_location, &finished,
                                 &file_changed, direction );
               if (rc == OK && rs > pl)
                  rs += net_change;
            }
         } else {
            if (g_status.control_break)
               rc = getkey( );
            /*
             * string not found     or   control break
             */
            error( WARNING, win->bottom_line,
                   g_status.control_break ? cb : find8 ); 
            finished = TRUE;
            rc = ERROR;
         }
      }
   } else {
      if ((replace_start = backward_boyer_moore_search( &wp )) != NULL &&
          !g_status.control_break) {
         rs = ptoul( replace_start );
         replace_and_display( &wp, replace_start, &finished, &file_changed,
                              direction );
      } else {
         /*
          * string not found
          */
         error( WARNING, win->bottom_line, find8 );
         finished = TRUE;
         rc = ERROR;
      }
      while (finished == FALSE) {
         update_line( &wp );
         if ((pattern_location = backward_boyer_moore_search( &wp )) != NULL &&
             !g_status.control_break) {
            pl = ptoul( pattern_location );
            if (rs == pl || (pl > rs && rs > pl - sub_len))
               finished = TRUE;
            else {
               rc = replace_and_display( &wp, pattern_location, &finished,
                                 &file_changed, direction );
               if (rc == OK && rs > pl)
                  rs += net_change;
            }
         } else {
            if (g_status.control_break)
               rc = getkey( );
            /*
             * string not found    or   control break
             */
            error( WARNING, win->bottom_line,
                   g_status.control_break ? cb : find8 );
            finished = TRUE;
            rc = ERROR;
         }
      }
   }

   if (mode.inflate_tabs)
      win->rcol = detab_adjust_rcol( win->cursor, win->rcol );

   dup_window_info( win, &wp );
   check_virtual_col( win, win->rcol, win->ccol );
   if (win->file_info->dirty != LOCAL && win->file_info->dirty != GLOBAL)
      show_curl_line( win );
   if (file_changed)
      win->file_info->modified = TRUE;
   make_ruler( window );
   show_ruler( window );
   show_ruler_pointer( window );
   return( rc );
}


/*
 * Name:    replace_and_display
 * Purpose: To make room for replacement string
 * Date:    June 5, 1991
 * Passed:  window:            pointer to current window
 *          pattern_location:  pointer to position of pattern in file
 *          finished:          stop replacing on this occurrence?
 *          modified:          skip this replacement?
 *          direction:         direction of search
 * Notes:   Show user where pattern_location is on screen if needed.
 *          Replace and display if needed.   Always ask the user if he
 *          wants to do wrapped replacing.
 */
int  replace_and_display( WINDOW *window, text_ptr pattern_location,
                          int *finished, int *modified, int direction )
{
register int rc;
file_infos *file;
register WINDOW *win;  /* put window pointer in a register */
int detab_rcol;

   win = window;
   file = win->file_info;
   rc = OK;
   if (g_status.wrapped) {
      rc = ask_wrap_replace( win, finished );
      g_status.wrapped = FALSE;
      show_search_message( CLR_SEARCH, g_display.mode_color );
   }
   if (rc == OK) {
      detab_rcol = win->rcol;
      find_adjust( win, pattern_location );
      make_ruler( win );
      show_ruler( win );
      show_ruler_pointer( win );
      xygoto( win->ccol, win->cline );
      if (file->dirty) {
         display_current_window( win );
         file->dirty = FALSE;
      } else
         show_curl_line( win );

      if (g_status.replace_flag == PROMPT && rc == OK) {
         show_line_col( win );
         rc = ask_replace( win, finished );
      }
      if (rc == OK) {
         do_replace( win, pattern_location, direction );
         *modified = TRUE;
         file->dirty = GLOBAL;
         show_changed_line( win );
         file->dirty = FALSE;
      }
      if (mode.inflate_tabs)
         win->rcol = detab_rcol;
   }
   return( rc );
}


/*
 * Name:    goto_top_file
 * Purpose: To move the cursor to the top of the file.
 * Date:    June 5, 1991
 * Passed:  window:  pointer to current window
 */
int  goto_top_file( WINDOW *window )
{
text_ptr next;          /* successive lines above the cursor */
register int i;
register WINDOW *win;   /* put window pointer in a register */

   win = window;
   entab_linebuff( );
   un_copy_line( win->cursor, win, TRUE );
   if (win->rline != win->cline - (win->top_line+win->ruler-1)) {
      next = cpf( win->file_info->start_text );
      for (i=win->cline; i>win->top_line+win->ruler; i--)
         next = find_next( next );
      win->cursor = next;
      win->rline = win->cline - (win->top_line+win->ruler-1);
      display_current_window( win );
   }
   sync( win );
   return( OK );
}


/*
 * Name:    goto_end_file
 * Purpose: To move the cursor to the end of the file.
 * Date:    June 5, 1991
 * Passed:  window:  pointer to current window
 */
int  goto_end_file( WINDOW *window )
{
text_ptr prev;
int i;
register int j;
file_infos *file;
register WINDOW *win;  /* put window pointer in a register */

   win = window;
   entab_linebuff( );
   un_copy_line( win->cursor, win, TRUE );
   file = win->file_info;
   if (file->length > win->rline + win->bottom_line - win->cline) {
      prev = cpb( file->end_text ) - 1;
      for (j=0,i=win->bottom_line; i>win->cline; i--, j++)
         prev = find_prev( prev );
      win->cursor = prev;
      win->rline = file->length - j + 1;
      display_current_window( win );
   }
   sync( win );
   return( OK );
}


/*
 * Name:    scan_forward
 * Purpose: To find the corresponding occurrence of target, ignoring
 *           embedded pairs of opp and target, searching forwards.
 * Date:    June 5, 1991
 * Passed:  start:  position of character to be paired
 *          opp:    the opposite to target
 *          target: the string to be found
 *          rc:     OK if found, ERROR otherwise
 * Returns: the location of the corresponding target in the text buffer
 * Notes:   Every 8,000 characters, check pointer forward.
 */
text_ptr scan_forward( text_ptr start, char opp, char target, int *rc )
{
text_ptr orig;
int count = 0;          /* number of unmatched opposites found */
register int check = 0;
register char c;

   rc = OK;
   orig = start = cpf( start );
   while ((c = *++start) && (c != CONTROL_Z)) {
      check++;
      if (opp == c)
         count++;
      else if (target == c) {
         if (count == 0)
            break;
         --count;
      }

      /*
       * after every 8,000 characters, check that pointer.
       */
      if (check > 8000) {
         start = cpf( start );
         check = 0;
      }
   }
   if (c == CONTROL_Z) {
      start = orig;
      *rc = ERROR;
   }
   return( start );
}


/*
 * Name:    scan_backward
 * Purpose: To find the corresponding occurrence of target, ignoring
 *           embedded pairs of opp and target, searching backwards.
 * Date:    June 5, 1991
 * Passed:  start:  position of character to be paired
 *          opp:    the opposite to target
 *          target: the string to be found
 *          rc:     OK if found, ERROR otherwise
 * Returns: the location of the corresponding target in the text buffer
 */
text_ptr scan_backward( text_ptr start, char opp, char target, int *rc )
{
text_ptr orig;
int count = 0;          /* number of unmatched opposites found */
register int check = 0;
register char c;

   *rc = OK;
   orig = start = cpb( start );
   while ((c = *--start) && (c != CONTROL_Z)) {
      check++;
      if (opp == c)
         count++;
      else if (target == c) {
         if (count == 0)
            break;
         --count;
      }

      /*
       * after every 8,000 characters, check that pointer.
       */
      if (check > 8000) {
         start = cpb( start );
         check = 0;
      }
   }
   if (c == CONTROL_Z) {
      start = orig;
      *rc = ERROR;
   }
   return( start );
}


/*
 * Name:    match_pair
 * Purpose: To find the corresponding pair to the character under the
 *           cursor.
 * Date:    June 5, 1991
 * Passed:  window:  pointer to current window
 * Notes:   Searching is very simple-minded, and does not cope with things
 *          like brackets embedded within quoted strings.
 */
int  match_pair( WINDOW *window )
{
text_ptr orig;  /* cursor location in text */
char c;
register WINDOW *win;  /* put window pointer in a register */
int rc;

   win = window;
   entab_linebuff( );
   un_copy_line( win->cursor, win, TRUE );
   /*
    * make sure the character under the cursor is one that has a
    *  matched pair
    */
   if ((unsigned)win->rcol >= linelen( win->cursor ))
      return( ERROR );
   win->cursor = cpf( win->cursor );
   orig = win->cursor + win->rcol;
   c = *orig;

   /*
    * find the matching pair
    */
   switch (c) {
      case '[':
         orig = scan_forward( orig, '[', ']', &rc );
         break;
      case '(':
         orig = scan_forward( orig, '(', ')', &rc );
         break;
      case '{':
         orig = scan_forward( orig, '{', '}', &rc );
         break;
      case ']':
         orig = scan_backward( orig, ']', '[', &rc );
         break;
      case ')':
         orig = scan_backward( orig, ')', '(', &rc );
         break;
      case '}':
         orig = scan_backward( orig, '}', '{', &rc );
         break;
      default :
         return( ERROR );
   }

   /*
    * now show the user what we have found
    */
   update_line( win );
   find_adjust( win, orig );
   if (!win->file_info->dirty)
      show_curl_line( win );
   make_ruler( win );
   show_ruler( win );
   return( rc );
}


/*
 * Name:    goto_line
 * Purpose: To move the cursor to a particular line in the file
 * Date:    June 5, 1991
 * Passed:  window:  pointer to current window
 */
int  goto_line( WINDOW *window )
{
long number;            /* line number selected */
long i;                 /* line counter */
char num_str[MAX_COLS]; /* line number as string */
text_ptr p;             /* used to scan through file counting lines */
register WINDOW *win;   /* put window pointer in a register */
int rc;

   win = window;
   entab_linebuff( );
   un_copy_line( win->cursor, win, TRUE );
   /*
    * find out where we are going
    */
   num_str[0] = '\0';
   /*
    * line number:
    */
   if (get_name( find11, win->bottom_line, num_str,
                 g_display.message_color ) != OK)
      return( ERROR );
   number = atol( num_str );

   if (number > 0  && number <= win->file_info->length) {
      update_line( win );
      p = win->cursor;
      i = win->rline;
      if (number < win->rline) {
         p = cpb( p );
         for (; i>number; i--)
            p = find_prev( p );
      } else {
         cpf( p );
         for (; i<number; i++)
            p = find_next( p );
      }
      find_adjust( win, p );
      if (!win->file_info->dirty)
         show_curl_line( win );
      rc = OK;
   } else {
      /*
       * out of range.  must be in the range 1 -
       */
      strcat( num_str, find12 );
      ltoa( win->file_info->length, num_str+25, 10 );
      error( WARNING, win->bottom_line, num_str );
      rc = ERROR;
   }
   return( rc );
}
