/*
 * Most of the tab routines were gathered into one file.  There is an
 *  assembly routine tdeasm.c that expands tabs.  That routine is in
 *  assembly to keep screen updates fairly fast.
 *
 * The basic detab and entab algorithms were supplied by Dave Regan,
 *   regan@jacobs.cs.orst.edu
 *
 * At school, I don't remember ever learning how to calculate tab positions.
 *   I guess my prof's thought no one would be stupid enough to write his
 *   own text editor.  Anyway, for more info on tabs see:
 *
 *       Brian W. Kernighan and P. J. Plauger, _Software Tools_, Addison-
 *       Wesley Publishing Company, Reading, Mass, 1976, pp 18-27.
 *       ISBN 0-20103669-X
 *
 * New editor name:  TDE, the Thomson-Davis Editor.
 * Author:           Frank Davis
 * Date:             June 5, 1991, version 1.0
 * Date:             July 29, 1991, version 1.1
 * Date:             October 5, 1991, version 1.2
 * Date:             January 20, 1992, version 1.3
 * Date:             February 17, 1992, version 1.4
 * Date:             April 1, 1992, version 1.5
 * Date:             June 5, 1992, version 2.0
 * Date:             October 31, 1992, version 2.1
 *
 * This program is released into the public domain, Frank Davis.
 *   You may distribute it freely.
 */

#include "tdestr.h"
#include "common.h"
#include "tdefunc.h"
#include "define.h"


/*
 * Name:    tab_key
 * Purpose: To make the necessary changes after the user types the tab key.
 * Date:    June 5, 1991
 * Passed:  window:  pointer to current window
 * Notes:   If in insert mode, then this function adds the required
 *           number of spaces in the file.
 *          If not in insert mode, then tab simply moves the cursor right
 *           the required distance.
 */
int  tab_key( WINDOW *window )
{
int spaces;      /* the spaces to move to the next tab stop */
char *source;    /* source for block move to make room for c */
char *dest;      /* destination for block move */
int pad, len;
register int rcol;
int old_bcol;
register WINDOW *win;  /* put window pointer in a register */
int rc;

   win  = window;
   if (*win->cursor == CONTROL_Z)
      return( OK );
   rcol = win->rcol;
   old_bcol = win->bcol;
   show_ruler_char( win );
   /*
    * work out the number of spaces to the next tab stop
    */
   if (mode.smart_tab)
      spaces = next_smart_tab( win );
   else
      spaces = mode.ltab_size - (rcol % mode.ltab_size);

   if (mode.insert && rcol + spaces < g_display.line_length) {
      copy_line( win->cursor, win->bottom_line );
      detab_linebuff( );
      /*
       * work out how many characters need to be inserted
       */
      len = linelen( g_status.line_buff );
      pad = rcol > len ? rcol - len : 0;
      if (g_status.line_buff[len] == CONTROL_Z)
         ++pad;
      if (len + pad + spaces >= g_display.line_length) {
         /*
          *  line too long to add
          */
         error( WARNING, win->bottom_line, ed1 );
         rc = ERROR;
         g_status.copied = FALSE;
      } else {
         if (pad > 0  || spaces > 0) {
            if (g_status.line_buff[len] == CONTROL_Z) {
               g_status.line_buff[len] = '\n';
               g_status.line_buff[len+1] = CONTROL_Z;
               ++win->file_info->length;
               show_size( window );
               --pad;
               ++len;
            }
            source = g_status.line_buff + rcol - pad;
            dest = source + pad + spaces;
            memmove( dest, source, len+pad-rcol+2 );

            /*
             * if padding was required, then put in the required spaces
             */
            memset( source, ' ', pad + spaces );
            entab_linebuff( );
         }

         win->file_info->dirty = GLOBAL;
         show_changed_line( win );
         rcol += spaces;
         win->ccol += spaces;
         rc = OK;
      }
   } else if (rcol + spaces <= g_display.line_length) {
      /*
       * advance the cursor without changing the text underneath
       */
      rcol += spaces;
      win->ccol += spaces;
      rc = OK;
   }
   check_virtual_col( win, rcol, win->ccol );
   if (old_bcol != win->bcol) {
      make_ruler( win );
      show_ruler( win );
   }
   return( rc );
}


/*
 * Name:    backtab
 * Purpose: To make the necessary changes after the user presses the backtab.
 * Date:    November 1, 1991
 * Passed:  window:  pointer to current window
 * Notes:   If in insert mode, then this function subs the required
 *           number of spaces in the file.
 *          If not in insert mode, then tab simply moves the cursor left
 *           the required distance.
 */
int  backtab( WINDOW *window )
{
int spaces;      /* the spaces to move to the next tab stop */
char *source;    /* source for block move to make room for c */
char *dest;      /* destination for block move */
int pad, len;
register int rcol;
int old_bcol;
register WINDOW *win;  /* put window pointer in a register */

   win  = window;
   rcol = win->rcol;
   if (*win->cursor == CONTROL_Z || win->rcol == 0)
      return( OK );
   old_bcol = win->bcol;
   show_ruler_char( win );

   /*
    * work out the number of spaces to the previous tab stop
    */
   if (mode.smart_tab)
      spaces = prev_smart_tab( win );
   else
      spaces = win->rcol % mode.ltab_size;

   if (spaces == 0)
      spaces = mode.ltab_size;
   copy_line( win->cursor, win->bottom_line );
   detab_linebuff( );
   len = linelen( g_status.line_buff );
   if (mode.insert && rcol - spaces < len) {
      pad = rcol > len ? rcol - len : 0;
      if (g_status.line_buff[len] == CONTROL_Z)
         ++pad;
      if (pad > 0  || spaces > 0) {
         if (g_status.line_buff[len] == CONTROL_Z) {
            g_status.line_buff[len] = '\n';
            g_status.line_buff[len+1] = CONTROL_Z;
            ++win->file_info->length;
            show_size( win );
            --pad;
            ++len;
         }
         /*
          * if padding was required, then put in the required spaces
          */
         if (pad > 0) {
            source = g_status.line_buff + rcol - pad;
            dest = source + pad;
            memmove( dest, source, pad+2 );
            memset( source, ' ', pad );
         }
         source = g_status.line_buff + rcol;
         dest = source - spaces;
         memmove( dest, source, len+pad-rcol+2 );
         entab_linebuff( );
      }

      win->file_info->dirty = GLOBAL;
      show_changed_line( win );
      rcol -= spaces;
      win->ccol -= spaces;
   } else {
      /*
       * move the cursor without changing the text underneath
       */
      rcol -= spaces;
      if (rcol < 0)
         rcol = 0;
      win->ccol -= spaces;
   }
   check_virtual_col( win, rcol, win->ccol );
   if (old_bcol != win->bcol) {
      make_ruler( win );
      show_ruler( win );
   }
   return( OK );
}


/*
 * Name:    next_smart_tab
 * Purpose: To find next smart tab
 * Date:    June 5, 1992
 * Passed:  window: pointer to the current window
 * Notes:   To find a smart tab 1) find the first non-blank line above the
 *            current line, 2) find the first non-blank character after
 *            column of the cursor.
 */
int  next_smart_tab( WINDOW *window )
{
register int spaces;    /* the spaces to move to the next tab stop */
text_ptr s;             /* pointer to text */

   /*
    * find first previous non-blank line above the cursor.
    */
   s = find_prev( cpb( window->cursor ) );
   while (s != NULL && is_line_blank( s ) )
      s = find_prev( s );
   if (s != NULL) {
      /*
       * if cursor is past the eol of the smart line, lets find the
       *   next fixed tab.
       */
      if (window->rcol >= find_end( s ))
         spaces = mode.ltab_size - (window->rcol % mode.ltab_size);
      else {

         s = detab_a_line( s );

         spaces = 0;
         s = cpf( s ) + window->rcol;

         /*
          * if we are on a word, find the end of it.
          */
         while (*s != ' '  &&  *s != '\n') {
            ++s;
            ++spaces;
         }

         /*
          * now find the start of the next word.
          */
         if (*s != '\n')
            while (*s == ' ') {
               ++s;
               ++spaces;
            }
      }
   } else
      spaces = mode.ltab_size - (window->rcol % mode.ltab_size);
   return( spaces );
}


/*
 * Name:    prev_smart_tab
 * Purpose: To find previous smart tab
 * Date:    June 5, 1992
 * Passed:  window: pointer to the current window
 * Notes:   To find a smart tab 1) find the first non-blank line above the
 *            current line, 2) find the first non-blank character before
 *            column of the cursor.
 *          there are several cases to consider:  1) the cursor is past the
 *            the end of the smart line, 2) the smart pointer is in the
 *            middle of a word, 3) there are no more words between the
 *            smart pointer and the beginning of the line.
 */
int  prev_smart_tab( WINDOW *window )
{
register int spaces;    /* the spaces to move to the next tab stop */
text_ptr s;             /* pointer to text */
unsigned int len;
int i;

   /*
    * find first previous non-blank line above the cursor, if it exists.
    */
   s = find_prev( cpb( window->cursor ) );
   while (s != NULL && is_line_blank( s ) )
      s = find_prev( s );

   if (s != NULL) {

      /*
       * if there are no words between the cursor and column 1 of the
       *   smart tab line, find previous fixed tab.
       */
      if (window->rcol < first_non_blank( s ))
         spaces = window->rcol % mode.ltab_size;
      else {
         s = cpf( s );
         len = find_end( s );

         /*
          * things are a little hairy w/ inflate_tabs.  in the main text
          *   buffer, lines are separated by <LF>.  if we expand tabs in the
          *   tabout buffer, there will not be a <LF> that represents the
          *   beginning of the next line.  if we are in inflate tab mode, lets
          *   slide the contents down one and insert a CONTROL_Z.  that way,
          *   when we go backwards looking for the beginning of words, we will
          *   run into CONTROL_Z when we hit the beginning of the line.
          */
         if (mode.inflate_tabs) {
            s = detab_a_line( s );
            _fmemmove( s+1, s, linelen( s ) + 2 );
            *s = CONTROL_Z;
            ++s;
         }


         /*
          * now, we need to figure the initial pointer and space.
          *   if the cursor is past the eol of the smart line, then
          *   set the smart pointer "s" to the end of line and "spaces" to
          *   the number of characters between the cursor and the eol
          *   of the smart line.  otherwise, set the smart pointer "s" to
          *   the column of the cursor and "spaces" to 0.
          */
         if (len < (unsigned)window->rcol) {
            s += len;
            spaces = window->rcol - len;
         } else {
            s += window->rcol;
            spaces = 0;
         }

         s = cpb( s );

         /*
          * if the cursor is past the eol of the smart line, find the start
          *   of the word that is at the eol.
          */
         if ((unsigned)window->rcol >= len) {

            /*
             * skip any space that has not been trimmed from the eol.
             */
            while (*(s-1) == ' ') {
               --s;
               ++spaces;
            }

            /*
             * now find the beginning of the first word at eol.
             */
            while (*(s-1) != ' ') {
               --s;
               ++spaces;
            }
         } else {
            /*
             * initially, lets keep a count of the number of non-spaces
             *   that we skip.  if the pointer happens to be in the middle
             *   of a word, then lets find the begining of the word.
             */
            i = 0;
            while (*s != ' ' && *s != '\n' && *s != CONTROL_Z) {
               --s;
               ++spaces;
               i++;
            }
            if (i > 1)
               /*
                * in the above while loop, we went one character too far when
                *   we found the beginning of the word under the pointer.
                *   if "i" is greater than 1, then the smart pointer "s"
                *   was somewhere in the middle of a word.  so, lets return
                *   the number of spaces needed to reach the beginning of
                *   the word from the current cursor position.
                */
               --spaces;
            else {
               /*
                * if we get this far, we have just found the first character
                *   of the first word of the smart pointer "s".  now, lets
                *   skip the spaces between the word we just found and the
                *   next previous word.
                * we may run into '\n' or ^Z if there are no previous words
                *   on the line.
                */
               while (*s == ' ') {
                  --s;
                  ++spaces;
               }

               /*
                * now, lets find the start of the word we just found.
                */
               while (*s != ' '  &&  *(s-1) != ' ' && *s != '\n'  &&
                      *s != CONTROL_Z) {
                  --s;
                  ++spaces;
               }

               /*
                * if *s == '\n' or *s == CONTROL_Z, then there were no
                *   previous words.  lets find the previous fixed tab.
                */
               if (*s == '\n' || *s == CONTROL_Z)
                  spaces = window->rcol % mode.ltab_size;
            }
         }
         if (spaces > window->rcol)
            spaces = window->rcol;
      }
   } else
      spaces = window->rcol % mode.ltab_size;

   /*
    * spaces cannot be negative.
    */
   if (spaces < 0)
      spaces = 0;
   return( spaces );
}


/*
 * Name:    entab
 * Purpose: To compress spaces to tabs
 * Date:    October 31, 1992
 * Passed:  s: pointer to current line
 *          break_on:  when set on, stop entabing when we reach 1st \n
 *                     when off, entab until we reach CONTROL_Z
 * Returns: pointer to tabout_buf
 * Notes:   the text_ptr can point to either the g_status.line_buff or
 *            a line in the main text buffer.  we need to know when
 *            to stop entabbing or we may have a run-away entab function.
 */
text_ptr entab( text_ptr s, int break_on )
{
int  tab_size;
int  last_col;
int  space;
register int col;
text_ptr to;

   tab_size = mode.ptab_size;
   to = g_status.tabout_buff;
   s = cpf( s );
   for (last_col=col=0; *s != CONTROL_Z; s++) {
      if (*s == '\n') {

         /*
          * when we reach the eol, compress trailing spaces to tabs.
          *   the un_copy_line function is responsible for trimming
          *   trailing space.
          */
         if (col != last_col) {
            while (last_col < col) {
               space = tab_size - last_col % tab_size;
               if (space <= 1) {
                  *to++ = ' ';
                  last_col++;
               } else if (last_col + space <= col) {
                  *to++ = '\t';
                  last_col += space;
               } else {
                  *to++ = ' ';
                  last_col++;
               }
            }
         }
         *to++ = *s;

         /*
          * if s is a pointer to a line in the main text buffer, then
          *   we need to stop entabbing when we get to EOL.  otherwise,
          *   we will entab the rest of the file.
          */
         if (break_on == TRUE)
            break;
         else
            last_col = col = 0;
      } else if (*s == '\t')
         col = col + tab_size - (col % tab_size);
      else if (*s == ' ')
         col++;
      else {
         if (col != last_col) {
            while (last_col < col) {
               space = tab_size - last_col % tab_size;

               /*
                * when space == 1, forget about emmitting a tab
                *   for 1 space.
                */
               if (space <= 1) {
                  *to++ = ' ';
                  last_col++;
               } else if (last_col + space <= col) {
                  *to++ = '\t';
                  last_col += space;
               } else {
                  *to++ = ' ';
                  last_col++;
               }
            }
         }
         last_col = col = col + 1;
         *to++ = *s;

         /*
          * when we see a quote character, stop entabbing.
          */
         if (*s == '\"' || *s == '\'') {
            while (*(s+1) != '\n'  &&  *(s+1) != CONTROL_Z)
              *to++ = *++s;
         }
      }
   }
   *to = CONTROL_Z;
   return( g_status.tabout_buff );
}


/*
 * Name:    detab_linebuff
 * Purpose: To expand tabs in line buffer so we can handle editing functions
 * Date:    October 31, 1992
 * Notes:   before we do any editing function that modifies text, let's
 *            expand any tabs to space.
 *          if inflate_tabs is FALSE, then nothing is done.
 */
void detab_linebuff( void )
{
int show_eol;

   if (mode.inflate_tabs  &&  g_status.copied) {
      show_eol = mode.show_eol;
      mode.show_eol = FALSE;
      tabout( g_status.line_buff );
      memmove( g_status.line_buff, g_status.tabout_buff,
                       linelen( g_status.tabout_buff ) + 2 );
      mode.show_eol = show_eol;
   }
}


/*
 * Name:    entab_linebuff
 * Purpose: To compress space in line buffer
 * Date:    October 31, 1992
 * Notes:   compress spaces back to tabs in the line buffer, if
 *             inflate_tabs == TRUE
 */
void entab_linebuff( void )
{
   if (mode.inflate_tabs  &&  g_status.copied) {
      entab( g_status.line_buff, FALSE );
      memmove( g_status.line_buff, g_status.tabout_buff,
                       find_CONTROL_Z( g_status.tabout_buff ) + 1 );
   }
}


/*
 * Name:    detab_a_line
 * Purpose: To inflate tabs in any line in the file
 * Date:    October 31, 1992
 * Returns: pointer to text or tabout_buff
 * Notes:   expand an arbitrary line in the file.
 */
text_ptr detab_a_line( text_ptr s )
{
int show_eol;

   if (mode.inflate_tabs) {
      show_eol = mode.show_eol;
      mode.show_eol = FALSE;
      s = tabout( s );
      mode.show_eol = show_eol;
   }
   return( s );
}


/*
 * Name:    detab_adjust_rcol
 * Purpose: given rcol in a line, find virtual column
 * Date:    October 31, 1992
 * Passed:  s:  string pointer
 * Notes:   without expanding tabs, calculate the display column according
 *            to current tab stop.
 */
int  detab_adjust_rcol( text_ptr s, int rcol )
{
register int col;

   s = cpf( s );
   for (col=0; rcol >= 0; rcol--,s++) {
      if (*s == '\t')
         col += (mode.ptab_size - (col % mode.ptab_size));
      else if (rcol > 0)
         col++;
   }
   return( col );
}


/*
 * Name:    entab_adjust_rcol
 * Purpose: given virtual rcol in a line, find real column
 * Date:    October 31, 1992
 * Passed:  s:  string pointer
 * Notes:   without expanding tabs, calculate which col the real cursor
 *            referencing.
 */
int  entab_adjust_rcol( text_ptr s, int rcol )
{
register int col;
register int last_col;

   s = cpf( s );
   for (last_col=col=0; col < rcol  &&  *s != CONTROL_Z  &&  *s != '\n'; s++) {
      if (*s != '\t')
         ++col;
      else
         col += (mode.ptab_size - (col % mode.ptab_size));
      if (col > rcol)
         break;
      ++last_col;
   }
   return( last_col );
}


/*
 * Name:    block_expand_tabs
 * Purpose: Expand tabs in a marked block.
 * Date:    June 5, 1991
 * Passed:  window:  pointer to current window
 * Notes:   Tabs are expanded using the current tab interval.
 *          Lines are checked to make sure they are not too long.
 */
int  block_expand_tabs( WINDOW *window )
{
int prompt_line;
int len;
int tab;
int tab_size;
int dirty;
register int spaces;
int net_change;
text_ptr p;                     /* pointer to block line */
file_infos *file;
WINDOW *sw, s_w;
long er;
int i;
char *b, *d, *lb;

   /*
    * make sure block is marked OK and that this is a LINE block
    */
   prompt_line = window->bottom_line;
   un_copy_line( window->cursor, window, TRUE );
   check_block( );
   if (g_status.marked == TRUE) {

      file  = g_status.marked_file;
      if (file->block_type != LINE) {
         /*
          * can only expand tabs in line blocks
          */
         error( WARNING, prompt_line, block20 );
         return( ERROR );
      }

      /*
       * set the command to word wrap so the un_copy_line function will
       * not display the lines while expanding.
       */
      g_status.command = WordWrap;

      /*
       * initialize everything
       */
      dirty = FALSE;
      tab_size = mode.ptab_size;
      sw = g_status.window_list;
      for (; ptoul( sw->file_info ) != ptoul( file );)
         sw = sw->next;
      dup_window_info( &s_w, sw );
      p  = cpf( file->block_start );
      er = file->block_er;
      lb = g_status.line_buff;
      s_w.rline = file->block_br;
      for (; s_w.rline <= er  &&  !g_status.control_break; s_w.rline++) {

         /*
          * use the line buffer to expand LINE blocks.
          */
         tab = FALSE;
         len = linelen( p );
         net_change = 0;
         g_status.copied = FALSE;

         copy_line( p, prompt_line );
         for (b=lb, i=1; *b != CONTROL_Z; b++) {

            /*
             * each line in the LINE block is copied to the g_status.line_buff.
             *  look at the text in the buffer and expand tabs.
             */
            if (*b == '\t') {
               tab = TRUE;
               spaces = i % tab_size;
               if (spaces)
                  spaces = tab_size - spaces;
               if (spaces) {
                  d = b + spaces;
                  memmove( d, b, linelen( b )+2 );
               }
               memset( b, ' ', spaces+1 );
               net_change += spaces;
               i += spaces + 1;
               b += spaces;
            } else
               i++;
         }

         /*
          * if any tabs were found, write g_status.line_buff to file.
          */
         if (tab) {
            un_copy_line( p, &s_w, TRUE );
            dirty = TRUE;
         }
         p = find_next( p );
      }

      /*
       * IMPORTANT:  we need to reset the copied flag because the cursor may
       * not necessarily be on the last line of the block.
       */
      g_status.copied = FALSE;
      if (dirty) {
         check_block( );
         file->dirty = GLOBAL;
      }
   }
   return( OK );
}


/*
 * Name:    block_compress_tabs
 * Purpose: Expand tabs in a marked block.
 * Date:    October 31, 1992
 * Passed:  window:  pointer to current window
 * Notes:   Tabs are compress using the current tab setting.
 */
int  block_compress_tabs( WINDOW *window )
{
register int col;
register int spaces;
int prompt_line;
int last_col;
int tab;
int tab_size;
int dirty;
text_ptr p;                     /* pointer to block line */
text_ptr from;                  /* line in main text buff being compressed */
file_infos *file;
WINDOW *sw, s_w;
long er;
char *to;
int  indent_only;

   /*
    * make sure block is marked OK and that this is a LINE block
    */
   prompt_line = window->bottom_line;
   entab_linebuff( );
   un_copy_line( window->cursor, window, TRUE );
   check_block( );
   if (g_status.marked == TRUE) {

      file  = g_status.marked_file;
      if (file->block_type != LINE) {
         /*
          * can only compress tabs in line blocks
          */
         error( WARNING, prompt_line, block26 );
         return( ERROR );
      }

      indent_only = g_status.command == BlockIndentTabs ? TRUE : FALSE;

      /*
       * set the command to word wrap so the un_copy_line function will
       * not display the lines while expanding.
       */
      g_status.command = WordWrap;

      /*
       * initialize everything
       */
      dirty = FALSE;
      tab_size = mode.ptab_size;
      sw = g_status.window_list;
      for (; ptoul( sw->file_info ) != ptoul( file );)
         sw = sw->next;
      dup_window_info( &s_w, sw );
      p  = cpf( file->block_start );
      er = file->block_er;
      s_w.rline = file->block_br;
      for (; s_w.rline <= er  &&  !g_status.control_break; s_w.rline++) {

         /*
          * use the line buffer to compress LINE blocks.
          */
         tab = FALSE;

         from = p;
         to   = g_status.line_buff;

         for (last_col=col=0; *from != CONTROL_Z; from++) {
            if (*from == '\n') {
               if (col != last_col) {
                  while (last_col < col) {
                     spaces = tab_size - last_col % tab_size;
                     if (spaces <= 1) {
                        *to++ = ' ';
                        last_col++;
                     } else if (last_col + spaces <= col) {
                        *to++ = '\t';
                        last_col += spaces;
                        tab = TRUE;
                     } else {
                        *to++ = ' ';
                        last_col++;
                     }
                  }
               }
               *to++ = *from;
               break;
            } else if (*from == '\t')
               col = col + tab_size - (col % tab_size);
            else if (*from == ' ')
               col++;
            else {
               if (col != last_col) {
                  while (last_col < col) {
                     spaces = tab_size - last_col % tab_size;
                     if (spaces <= 1) {
                        *to++ = ' ';
                        last_col++;
                     } else if (last_col + spaces <= col) {
                        *to++ = '\t';
                        last_col += spaces;
                        tab = TRUE;
                     } else {
                        *to++ = ' ';
                        last_col++;
                     }
                  }
               }
               last_col = col = col + 1;
               *to++ = *from;
               if (*from == '\"' || *from == '\''  || indent_only) {
                  while (*(from+1) != '\n'  &&  *(from+1) != CONTROL_Z)
                     *to++ = *++from;
               }
            }
         }
         *to = CONTROL_Z;

         /*
          * if any tabs were found, write g_status.line_buff to file.
          */
         if (tab) {
            g_status.copied = TRUE;
            un_copy_line( p, &s_w, TRUE );
            dirty = TRUE;
         }
         p = find_next( p );
      }

      /*
       * IMPORTANT:  we need to reset the copied flag because the cursor may
       * not necessarily be on the last line of the block.
       */
      g_status.copied = FALSE;
      if (dirty) {
         check_block( );
         file->dirty = GLOBAL;
      }
   }
   return( OK );
}
